/** TITLE: XMLDeserialize.java (JSX 0.2).
	*        -------------------
	*
	* Creates the object described by an XML document.

	* To be serializable, classes and fields must be public.
	* For deserialization, a no-argument constructor is also necessary.

	* Incompatiable with JSX0.1; the redesign allows greater reuse, and a more
	* elegant DTD (DTD code is not yet complete).

	* Note that a basic limitation of Java is that an object cannot be created
	* unless the JVM has access to the class file defining it.  That is, Java
	* cannot *define* classes at runtime; it can only instantiate pre-defined
	* classes.
	* However, a class loader could be written to "read" in class files which
	* do not actually exist; thus giving Java the ability to define new
	* classes dynamically.

	* Please email me: for help, to suggest features, or to report bugs.
		bren@csse.monash.edu.au  <G. Brendan Macmillan>
	
	**/


//package JSX;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;	//for Hashtable

public class XMLDeserialize {

	private static final boolean DEBUG = false;
	//private static final boolean DEBUG = true;
	private ParserXML p;


	
	/**--------------------------------------------------------------------------
		* API
		* ===
		**/
	public XMLDeserialize(Reader in) {	
		p = new ParserXML(in);
	}
	public Object deserialize() throws Exception {
		ParserXML.Tag tag = p.readTag();
		if (tag==null) return null;		//no more objects left
		if (!tag.start)
			throw new ParserXML.ExceptionXML("Expected start tag, got \"</" +tag.name+ ">\"");
		return createObject(tag.name, null);	//no parent
	}


	/**--------------------------------------------------------------------------
		* Object (Element maps-to Object)
		* ------
		* Returns the deserialized object (caller needs to cast at runtime)
		* The object created is "me".
		
		* USAGE: the open tag is assumed to have been just read

		* LOGIC: A new object is instantiated: "me".  Attr and elements arex
		* added to me.  This new object is then added to its parent.  NB: that
		* the name for me (given by parents) is only discoverable by reading the
		* attr; which is done in "addAttr" (second arg is boolean for isArray)

		* A note on 'null': a null object may still have a name

		* NOTE: tag problem for lengths - not actually a field.
		**/
	private Object createObject(String elementName, Object parent)
			throws Exception {
		Object me;
	//mapping from my conventions to actual classes...
		if (elementName.equals(XMLSerialize.NULL_TOKEN))
			me = null;
		else if (elementName.startsWith(XMLSerialize.ARRAY_TOKEN)) {
			me = createArray(elementName, parent);	//need length here...
			return me;		//better to do all else internally (inc name & parent)
			//NB: it *must* read right up to this close of this tag
		}
		else
			me = Class.forName(elementName).newInstance();	//only newInstance in file
		//NB: an array may also have a name!
		ParserXML.Attr name = addAttr(me, false);	//adds in all attr and gets NAME_TOKEN
			//inside: if me is an array, don't add no fields!

		if (DEBUG) System.err.println("Made class \"" +elementName+ "\"");
			//needs to be after "me" is created...
			//but array needs the length before can be created - circular dependancy
			//best to fix while rationalizing the fields.
		//or, allow arrays to handle their attributes separately - this is correct.

		if (parent!=null) {		//root doesn't have a parent
			Field f = parent.getClass().getField(name.value);
			f.set(parent, me);
		}		//awkward meanings: name==null; and name.emptyTag
		if (DEBUG)
			System.err.println("Got an: "+  name);

		if (!name.emptyTag) {		//nice! (ensure 
			if (me!=null && me.getClass() == Vector.class)
				addVectorElements((Vector) me);	//(could poly it.)
			else if (me!=null && me.getClass() == Hashtable.class)
				addHashtableElements((Hashtable) me);
				//name.length;		//HACK - message passing
			else
				addTags(me);
		}
		return me;
	}


	/**--------------------------------------------------------------------------t
		* Create Array
		* ------------
		* Create one Array object.
		* Note that the whole name must be parsed to know the component type,
		* which may be: (ARRAY_TOKEN)*Type

		* Essentially, maps (ARRAY_TOKEN)*Type -> Type([])*
		* It counts the ARRAY_TOKENS, extracts the type, then:
		*  Array.newInstance(Class.getClass(baseType), new int[count])

		* NB: we avoid the complex disambiguation constraint on tok, by name choice
		**/
	private Object createArray(String tagName, Object parent) throws Exception {
		//System.err.println("ARRAY DESERIALIZATION NOT YET IMPLEMENTED (properly)");
		//As I say, gentle reader, there is lots of awful hacking here - but think!
		//- that is just how Michael Angelo brought forth David.  He hacked at
		//several blocks of stone, gaining understanding with each, before
		//completing the final masterpiece.  I am at the still at the first block,
		//and I've learnt a lot about the shape of the problem from it!
		// - Brendan "hacker or sculpter? you decide!" Macmillan

		//NOTE!! need this here, so we know length
		ParserXML.Attr attr = addAttr(null, true); //to get name and length
			//attr==null; attr.emptyTag; attr.length
			//NB!! addAttr should *only* read the name and length: there is no point
			//in reading the element values themselves, because the array itself has
			//not yet been created.
			//This implies that the length *must* be given before any element values.
			//It doesn't logically preclude the name to be given at any point, but
			//the present organisation of the program expects it also to be given
			//before any of the element values.

			//THUS, we need
			// (1). addAttr to return after seeing "length"; and
			// (2). to call addAttr a second time, treating attr as indices.

			//MORE problems!  addAttr needs to behave differently depending or
			//whether it is of primitive or of reference base type.  If primitive,
			//it returns immediately on reading the length;  if reference, it
			//swollows up the ">", at the end of the tag.
			//SOLN: because an array of reference type can never contain any
			//attr but name and length, we can additionally read these in, in a
			//hack.  All this must be rewritten cleanly, now that all (I think)
			//consequences of the data-binding are apparent - or perhaps a rewrite
			//of the data-binding might be more appropriate  ;-)
		int dim = 0;
		int index=0;
		String tok = XMLSerialize.ARRAY_TOKEN;
		int skip = tok.length();
		//(1). take off the leading "ArrayOf-"'s (by incrementing an index for eff)
		while (tagName.startsWith(tok, index)) { //could be > eff if assume tok qual
			dim++;														//and just check first char...
			index+=skip;
		}
		if (DEBUG) System.err.println("dim= " +dim);
		String type = tagName.substring(index);	//get the type
		if (DEBUG) System.err.println("type= " +type);
		Object me = null;
		try {
			int[] dimsArray = new int[dim];
			dimsArray[0] = attr.length;		//NOTE use of length parsed from attributes
			Class comp;
			//shockingly inefficient!  This is probably why they used the 'L'
			if (type.equals("int")) comp = int.class;
			else if (type.equals("double")) comp = double.class;
			else if (type.equals("boolean")) comp = boolean.class;
			else if (type.equals("char")) comp = char.class;
			else if (type.equals("byte")) comp = byte.class;
			else if (type.equals("short")) comp = short.class;
			else if (type.equals("long")) comp = long.class;
			else if (type.equals("float")) comp = float.class;
			else comp = Class.forName(type);
			me = Array.newInstance(comp, dimsArray);
		} catch (Exception e) {
			System.err.println("Illegal type as Array base component, it seems; "+e);
		}
		if (DEBUG) System.err.println("type= " +me.getClass().getName());
		if (DEBUG) System.err.println("len= " +Array.getLength(me));

		ParserXML.Tag tag;
		if (dim==1 && me.getClass().getComponentType().isPrimitive()) {
			//read all the following attrs into the array
			attr = addAttr(me, true); //NB: length is not needed again by this route
			if (DEBUG) System.err.println("after setting primitive array, saw: " + attr);
		}	//attr (for primitive base type)
		else {
			//do the contained objects (as does Vector)
			//NOTE  the recursion runs through create object, which tests isArray,
				//so as to automatically bottom out.
				//AND WHAT A HACK IT IS!
			ParserXML.Attr end = p.readAttr(); //slurp up the left-over '>' after length read.
			end.length = attr.length;
			if (DEBUG) System.err.println("after starting reference array, saw: " + end);
				//NB: this implies that the name (if present) *must* precede length
			for (int i=0; i<end.length; i++) {
				Object obj;
		 		tag=p.readTag();
				if (tag==null) throw new Exception("Premature EOF");
				if (!tag.start)
					throw new Exception("Expected "+end.length+" elements in array; read only"+i);
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					obj = deserializeWrapper(tag);
				else
					obj = createObject(tag.name, null);	//null=orphan object
				if (DEBUG) System.err.println("Adding {"+obj+"} to Array");
				Array.set(me, i, obj);
			}	//for loop
		}	//tags (for refernce base type - array or class)
		if (!attr.emptyTag) {		//nice!
			tag=p.readTag();
			if (tag==null) throw new Exception("Premature EOF");
			if (tag.start)
				throw new Exception("End of array");
			if (!tag.name.equals(tagName))
				throw new ParserXML.ExceptionXML(
					"</"+tagName+">","</"+tag.name+">");
		}
			//if get here, all is OK!

		if (parent!=null) {		//root doesn't have a parent CODE DUP
			Field f = parent.getClass().getField(attr.value);
			f.set(parent, me);
		}		//misuse as message gives awkward meanings:

		return me;
	}


	/** convert Array name string from:
		*		[*		//any number of '[' (including 0)
		*		(B|C|D|F|I|J|S|Z) |		//one of these represeting a primitive type
		*   (L[^;]*;)		//or the name of the class, in between 'L' and ';'
		**/
/*
	private static String getArrayType(Class clazz) {
		String a = "";	//will hold result
		String name = clazz.getName();
		if (DEBUG) System.err.println("Array type = \""+name+"\"");
		int i;
		for (i=0; name.charAt(i)=='['; i++)
			a += ARRAY_TOKEN;
		switch (name.charAt(i++)) {
			case 'B': a += "byte";
				break;
			case 'C': a += "char";
				break;
			case 'D': a += "double";
				break;
			case 'F': a += "float";
				break;
			case 'I': a += "int";
				break;
			case 'J': a += "long";
				break;
			case 'S': a += "short";
				break;
			case 'Z': a += "boolean";
				break;
			case 'L': a += name.substring(i, name.lastIndexOf(';'));
						//could also be name.length()-1
			break;
		}
		// NB: in a multi-dimensional array, Class.getComponentType() returns
		// the next one further in (which may also be an array).
		return a;
	}
*/



	/**--------------------------------------------------------------------------t
		* Attributes (map to primitives and String)
		* ----------
		* Only: int and String
		*	Should be: byte,char,short,int,long,float,double and String.
		*
		* BIZARRE: Could do something odd for arrays: take the object only as
		* indicating the exact type of the array, and in here, when read the
		* length, actually create it.
		*	And yet it makes sense in a way.
		*
		* or, return a proper "attrMsg", which includes the length as info.
		* This is wha should be returned, anyway - not this hacked use of Attr.
		**/
	private ParserXML.Attr addAttr(Object o, boolean isArray) throws Exception {
		int length = -1;	//to flag if not set
		ParserXML.Attr attr;
		ParserXML.Attr name = null;
		while ( !(attr=p.readAttr()).isEnd ) {
			if (DEBUG) System.err.print(attr+", ");
			if (attr.name.equals(XMLSerialize.NAME_TOKEN)) { //if name of this class
				if (name!=null)
					throw new ParserXML.ExceptionXML("just one name field");
				name = attr;
				continue;	//next attr
			}
				//ignore all fields of array, but for the name
			if (isArray) {
				if (attr.name.equals(XMLSerialize.LENGTH_TOKEN)) {
					length = Integer.parseInt(attr.value);
					break;		//break off part way through (will this work?)
				}
				int index = Integer.parseInt(attr.name);	//we know it is an int
				if (DEBUG) System.err.println(attr);
				Object value = getValue(o.getClass().getComponentType(), attr);		//hmmmph!  Seems quite simple
				Array.set(o, index, value);
			}
			else {
				Field f = o.getClass().getField(attr.name);
				Class fc = f.getType();			//type specified by java, not by xml
				if (DEBUG) System.err.println("Made Field \""+fc.getName()+"\"");
				f.set(o, getValue(fc, attr));
			}
		}
			//combine info from the name and final attr returned.  Awkward.
		attr.length = length;
		if (name==null) return attr;	//simple.  No name to return
			//else, combine info from both (name and last attr)
		name.isEnd = attr.isEnd;
		name.emptyTag = attr.emptyTag;
		name.length = attr.length;
		return name;
	}

	/** It may be more efficient to not reuse this code.
		* when used for an attribute, it creates an Object needlessly;
		* however, such an object is required for Hashtable
		**/
	private Object getValue(Class fc, ParserXML.Attr attr) throws Exception {
		if (DEBUG) System.err.println("getting value of " +fc);
		if (fc.isPrimitive()) {     //PRIMITIVE
			if (fc==int.class) return new Integer(attr.value);	//most likely first
			else if (fc==double.class) return new Double(attr.value); //second pr
			else if (fc==boolean.class) {
				if (attr.value.equals("false"))
					return new Boolean (false);
				else if (attr.value.equals("true"))
					return new Boolean (true);
				else
					throw new Exception("Boolean must be 'true' or 'false'; not '"
						+attr.value+"'");
			}
			else if (fc==byte.class) return new Byte(attr.value);
			else if (fc==char.class) {
				if (attr.value.length()!=1) throw new Exception("Character data must be only one character long");
				return new Character(attr.value.charAt(0));
			}
			else if (fc==short.class) return new Short(attr.value);
			//else if (fc==int.class) return new Integer(attr.value));
			else if (fc==long.class) return new Long(attr.value);
			else if (fc==float.class) return new Float(attr.value);
			else throw new ParserXML.ExceptionXML(
				"Unimplemented primitive: \"" +fc.getName()+ "\"");
		}
		else {                      //CLASS
			if (fc==String.class) return decodeXML(attr.value);
			else throw new ParserXML.ExceptionXML(
				"'"+attr.name+"' is not primitive or String");
		}
	}



	/**--------------------------------------------------------------------------
		* Elements
		* --------
		* Handles just tags (no attrs)
		* CONSIDER: is there an elegant way to reuse this code in Hashtable?
		* This is where the end tag is read
		* Perhaps pass in the open tag explicity, rather than (inefficiently)
		* recalling it from the object itself.  Why?  Because this doesn't work
		* for null objects.
		**/
	private void addTags(Object me) throws Exception {
		ParserXML.Tag tag;
		while ( ((tag=p.readTag())!=null) && (tag.start))	
			createObject(tag.name, me);	//add this new tag to me
		if (tag==null)
			throw new ParserXML.ExceptionXML("Premature EOF");
	
		if (me==null) {		//dangling else!
			if (!tag.name.equals(XMLSerialize.NULL_TOKEN))	//special case of null
				throw new ParserXML.ExceptionXML(
					"</"+XMLSerialize.NULL_TOKEN+">","</"+tag.name+">");
		}
		else
			if (!tag.name.equals(me.getClass().getName()))
				throw new ParserXML.ExceptionXML(
					"</"+me.getClass().getName()+">","</"+tag.name+">");
	}


	/**--------------------------------------------------------------------------
		* Vector (container object)
		* ------
		* Note: Vector object already created when we get here.
		**/
	private void addVectorElements(Vector v) throws Exception {
		ParserXML.Tag tag;
		while (true) {
			Object obj;
			tag = p.readTag();
			if (tag==null) throw new Exception("Premature EOF");
			if (!tag.start) return;	//done!  caller will check if a correct endtag
			if (tag.name.equals("java.lang.String")	//bad: doubles up
					|| tag.name.equals("java.lang.Integer")
					|| tag.name.equals("java.lang.Double")
					|| tag.name.equals("java.lang.Boolean")
					|| tag.name.equals("java.lang.Character")
					|| tag.name.equals("java.lang.Byte")
					|| tag.name.equals("java.lang.Short")
					|| tag.name.equals("java.lang.Long")
					|| tag.name.equals("java.lang.Float")
				)
				obj = deserializeWrapper(tag);
			else
				obj = createObject(tag.name, null);	//null=orphan object
			if (DEBUG) System.err.println("Adding {"+obj+"} to Vector");
			v.addElement(obj);
		}
		//if (DEBUG) System.err.println("Vector: \n" +v);
	}

	/**--------------------------------------------------------------------------
		* Hashtable
		* ---------
		* An container object.
		* Handles just tags (no attrs)
		**/
	private void addHashtableElements(Hashtable h) throws Exception {
		ParserXML.Tag tag;
		while (true) {
			Object[] objs = {null, null};
			for (int i=0; i<objs.length; i++) {		//{key, value}
				tag = p.readTag();
				if (tag==null) throw new Exception("Premature EOF");
				if (!tag.start)
					if (i==0) return; //end of hashtable
					else throw new Exception("incomplete Hashtable entry");
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					objs[i] = deserializeWrapper(tag);
				else
					objs[i] = createObject(tag.name, null);	//null=orphan object
			}
			if (DEBUG) System.err.println("Putting {"+objs[0]+", "+objs[1]+"}");
			h.put(objs[0], objs[1]);
		}
		//if (DEBUG) System.err.println("Hashtable: \n"+h);
	}


	/**--------------------------------------------------------------------------
		* Wrapper
		* -------
		* Handles the attrs.
		* Need to refactor the test for wrapper type
		**/
	private Object deserializeWrapper(ParserXML.Tag startTag) throws Exception {
		ParserXML.Attr attr, end;
		ParserXML.Tag tag;
		attr = p.readAttr();
		if (!attr.name.equals("valueOf")) 
			throw new ParserXML.ExceptionXML("valueOf expected");
		if (!(end = p.readAttr()).isEnd)
			throw new ParserXML.ExceptionXML("\">\" or \"/>\" expected");
		if (!end.emptyTag) {
			tag = p.readTag();
			if (!tag.name.equals(startTag.name) || tag.start)
				throw new ParserXML.ExceptionXML("</"+tag.name+"> expected");
		}
		//IDEA: could make up the expected tag, and use Tag.equals() to compare
		if (DEBUG) System.err.println("String read in: " + attr.value);
		if (startTag.name.equals("java.lang.String"))
			return decodeXML((String)attr.value);	//not attr.name! Don't forget to decode!
		Class c = Class.forName(startTag.name);
		try {
			Field f = c.getField("TYPE");
			return getValue((Class)f.get(c), attr);
		} catch (NoSuchFieldException e) {
			System.err.println("'" +c+ "' is not a wrapper class: "+e);
		}
		throw new Exception("Shouldn't get here");
	}


	/**--------------------------------------------------------------------------
		* decodeXML
		* ---------
		**/
	private String decodeXML(String in) {
		String a = "";
		int i;
		try {
			for (i=0; i<in.length(); i++) {
				char c = in.charAt(i);
				if (c=='&') {				//if the escape character
					i++;
					if (in.startsWith("amp;", i))       {a+='&';  i+=3;} //NB: loop's x++
					else if (in.startsWith("lt;", i))   {a+='<';  i+=2;}
					else if (in.startsWith("gt;", i))   {a+='>';  i+=2;}
					else if (in.startsWith("apos;", i)) {a+='\''; i+=4;}
					else if (in.startsWith("quot;", i)) {a+='"';  i+=4;}
				}
				else
					a += c;
			}
		} catch (IndexOutOfBoundsException e) {
			System.err.println( "Seems to be an incomplete escaped entity: "
				+in.substring(in.lastIndexOf('&')) );
		}
		return a;
	}


	/**--------------------------------------------------------------------------
		* Test class
		* ----------
		**/
	public static class Test {
		public static void main(String args[]) throws Exception {
			XMLDeserialize d = new XMLDeserialize(new InputStreamReader(System.in));
			Object o;
			while ( (o = d.deserialize()) != null) {	//can loop for streamed XML
			/*
				System.err.println("Printing created object:");
				System.err.println();
				System.err.println(o);
				System.err.println();
				System.err.println("(Re)serializing created object:");
				System.err.println();
			*/
				XMLSerialize.serialize(o);
			}
		}
	}
}
