/** TITLE: XMLSerialize.java (JSX 0.2).
	*        -----------------
	*
	* Pass in an objet, and it writes out a serialized version, in XML format.

	* Mapping Java to XML:
	*		Primitives map to attributes.
	*		References map to elements:
	*			The element name is the *runtime class*.  eg <java.util.Hashtable>
	*			A NAME_TOKEN attribute is the name of the field (if has a name)
	*				NB: NAME_TOKEN is an illegal Java field, so no namespace conflict.

	* Logic flow:
	*		For each class: (serializeObject)
	*			write all fields as attributes - if primitive (serializeFields)
	*			write all fields as elements - if reference (serializeReferences)
	*				-> recurse

	* To do:
	*  - ignore fields marked "transient".
	*  - write out a DTD (the new runtime class tags makes this more elegant)
	*	 - write alternative to StreamTokenizer for handling \n in quotes

	* Please email me: for help, to suggest features, or to report bugs.
		bren@csse.monash.edu.au  <G. Brendan Macmillan>

	* NOTE: Newlines within strings are outputted *exactly* as they are stored
	*	in the string.  This might not be what you want, when moving across DOS
	* (/n/r) and UNIX (/n).

	* TODO: nice, informative error messages when the field is not found
	
	**/

//package JSX;
import java.lang.reflect.*;
import java.io.*;		//for writeDTD
import java.util.*;	//for Hashtable


public class XMLSerialize {

	public static PrintWriter out;

	static final String NAME_TOKEN = "obj-name";	//"-" legal attr; illegal field
	//static final String NAME = "is-the";	//"-" legal attr; illegal field
	//static final String NAME = "class-of";	//opposite of "instanceof"
	//static final String NAME = "run-type-of";	//opposite of "instanceof"
		//there is an issue with compatibility here!  I keep changing it, and old
		//serializations are no longer compatible with it.
	static final String ARRAY_TOKEN = "ArrayOf-";
	//static final String ARRAY_TOKEN = "A-";
	//static final String ARRAY_TOKEN = "[]";	//not allowed as XML names!
	static final String LENGTH_TOKEN = "length";
	static final String NULL_TOKEN = "null";

	static final String SPACER = "  ";		//\t
	static final boolean DEBUG = false;		//not actually used as yet
	//static final boolean DEBUG = true;		//not actually used as yet

	static final int NOT_SERIALIZED =
		Modifier.FINAL | Modifier.STATIC | Modifier.TRANSIENT;


	/**--------------------------------------------------------------------------
		* API
		* ===
		**/
	public static void serialize(Object o) {
		if (out==null)		//else keeps the old one for streamed serialization
			out = new PrintWriter(new OutputStreamWriter(System.out), true);
		serializeObject(o, null, "");
		out.println();
	}
		
	public static void serialize(Object o, PrintWriter newOut) {
		out = newOut;			//is there a nice equivalent to "this.out = out;" ?
		serializeObject(o, null, "");
		out.println();
	}


	/**--------------------------------------------------------------------------
		* Object (map to an Element)
		* ------
		* tag (name of element) is always the runtime class (type) of the object.
		* SUGGEST: write open and close tags separately; but problem of attributes
		*  check: Wrapper classes here (String, primitive wrappers)
		*  check: container classes (like Hashtable)
		**/
	private static void serializeObject(Object o, String fieldName, String indent) {
		if (o==null)	{ //an array object can also be null...
			printOpenTag(NULL_TOKEN, indent);
			out.println("/>");		//good to have control of the newline here
			return;	//a null object can have no children
		}
		try {
			if (!(o instanceof Serializable))	//mark such fields "transient"
				throw new NotSerializableException(o.getClass().getName());
		} catch (Exception e) { System.err.println(e); return; }
		String runTimeClass = o.getClass().getName();
		Class clazz = o.getClass();

//FACTOR this out into a separate method, so we can recurse nicely.
		if (clazz.isArray()) {
			serializeArrayReferences(o, fieldName, indent);
		}
		else if (clazz==Vector.class) {
			printOpenTag(runTimeClass, indent);
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
			if (((Vector)o).size() > 0) {
				out.println(">");		//move into ParserXML?
				serializeVectorReferences((Vector)o, indent);	//could poly
				printCloseTag(runTimeClass, indent);
			} else
				out.println("/>");		//empty tag
		}
		else if (clazz==Hashtable.class) {
			printOpenTag(runTimeClass, indent);
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
			if (((Hashtable)o).size() > 0) {	//same as Vector
				out.println(">");		//move into ParserXML?
				serializeHashtableReferences((Hashtable)o, indent);	//poly out; nicer
				printCloseTag(runTimeClass, indent);
			} else
				out.println("/>");		//empty tag
		}

		else {
			printOpenTag(runTimeClass, indent);
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
			serializeFields(o, indent);
			if (hasChildren(o)) {
				out.println(">");		//move into ParserXML?
				serializeReferences(o, indent);
				printCloseTag(runTimeClass, indent);
			} else
				out.println("/>");		//empty tag
		}

	}

	/**--------------------------------------------------------------------------
		* I wonder if these tag factorings should be in ParserXML?  It's not
		* parsing - quite the opposite - but it is at the same level as the parser.
		* In the parlace of Compilers, it is "code generation", I guess
		**/
	private static void printOpenTag(String tag, String indent) {
		out.print(indent + "<" +tag);
	}
	private static void printCloseTag(String tag, String indent) {
		out.println(indent + "</" +tag+ ">");
		//out.println();	//nicer formatting	hmmm... also used by Wrapper...
	}


	/**--------------------------------------------------------------------------
		*	serializeArrayReferences
		*	------------------------
		* Can contain primitives (no wrapper) or objects
		* If called on non-array, it serializes as an Object (needed for recursion)
		**/
	private static void serializeArrayReferences(Object o, String fieldName, String indent) {
		Class clazz;
		if (o==null || !( (clazz=o.getClass()).isArray() )) {
			serializeWrapperOrObject(o, indent);
			return;		//bottoms out.
		}
		String runTimeClass = o.getClass().getName();	//redundancy here (see above)
		//out.println(indent + "ARRAYS NOT YET IMPLEMENTED");
		runTimeClass = getArrayType(clazz);
		printOpenTag(runTimeClass, indent); //map "[" to "ArrayOf-" etc
		if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
		int length  = Array.getLength(o);
		printAttr(LENGTH_TOKEN, length+"");
		if (length==0)		//if o.length==0
			out.println("/>");		//empty tag
		else {
				//if primitive, put it here?
			//if *last* array (ie array of non-array), and also of a primitive.
			//well, to get here, must be an array...
			//...so check check to see if it as array of a primitive
			if ( clazz.getComponentType().isPrimitive() ) {
				for (int i=0; i<length; i++) {
					out.println();
					out.print(indent);
					printAttr(i+"", Array.get(o,i)+"");
				}
				out.println("/>");		//*cannot* contain other tags. (into ParserXML?)
			}
			else {	//ie component is a reference type (another array or an object)
				out.println(">");		//move into ParserXML?
				for (int i=0; i<length; i++)		//actually a recursion
					serializeArrayReferences(Array.get(o, i), null, indent+SPACER);
					//closeTag is inside
				printCloseTag(runTimeClass, indent);
			}
		}
	}

	/** convert Array name string from:
		*		[*		//any number of '[' (including 0)
		*		(B|C|D|F|I|J|S|Z) |		//one of these represeting a primitive type
		*   (L[^;]*;)		//or the name of the class, in between 'L' and ';'
		**/
	private static String getArrayType(Class clazz) {
		String a = "";	//will hold result
		String name = clazz.getName();
		if (DEBUG) System.err.println("Array type = \""+name+"\"");
		int i;
		for (i=0; name.charAt(i)=='['; i++)
			a += ARRAY_TOKEN;
		switch (name.charAt(i++)) {
			case 'B': a += "byte";
				break;
			case 'C': a += "char";
				break;
			case 'D': a += "double";
				break;
			case 'F': a += "float";
				break;
			case 'I': a += "int";
				break;
			case 'J': a += "long";
				break;
			case 'S': a += "short";
				break;
			case 'Z': a += "boolean";
				break;
			case 'L': a += name.substring(i, name.lastIndexOf(';'));
						//could also be name.length()-1
			break;
		}
		// NB: in a multi-dimensional array, Class.getComponentType() returns
		// the next one further in (which may also be an array).
		return a;
	}

	/**--------------------------------------------------------------------------
		*	hasChildren
		*	-----------
		*	An inefficient but obvious prototype: better to factor out field
		* handlers (ie ref, primintive and this one), to farm out into appropriate
		* arrays for each.
		*	"Children" def: existing (ie non-null) classes (ie non-primitive).
		**/
	private static boolean hasChildren(Object o) {
		try {
			Field[] fields = o.getClass().getFields();
			for (int i=0; i<fields.length; i++) {
				Field f = fields[i];
				if ((f.getModifiers() & NOT_SERIALIZED) != 0)
					continue;
				if (!f.getType().isPrimitive() && f.getType()!=String.class) 
					if (f.get(o)!=null)
						return true;
			}
			return false;		//found no non-primitive (includes String),
				//non-null objects within it	
		}
		catch (Exception e) {System.err.println(e); System.exit(0);}
		return false;	//unreachable code
	}


	/**--------------------------------------------------------------------------
		* Fields (map to Attributes)
		* ------
		* SUGGEST: abstracting a "printAttr(String name, String value)" method
		**/
	private static void serializeFields(Object o, String indent) {
		if (o==null) return;
		Field[] fields = o.getClass().getFields();
			//Field[] fields = o.getClass().getDeclaredFields();
			//getDeclaredFeilds doesn't return inherited fields.
		try {
			for (int i=0; i<fields.length; i++) {
				Field f = fields[i];
				if ((f.getModifiers() & NOT_SERIALIZED) != 0)
					continue;	//skip final and static fields
				Class type = f.getType();
						//how bad would it be to treat these two following cases the same?
						//ie. and *always* do encodeXML() automatically, in the ParserXML
				if (type.isPrimitive())
				{
					out.println();							//Attribute indenting
					out.print(indent);		//indent each attribute
					printAttr(f.getName(), f.get(o)+"");
				}
				else if (type==String.class)	//.getName().equals("java.lang.String"))
				{
					out.println();							//Attribute indenting
					out.print(indent);		//indent each attribute
					printAttr(f.getName(), encodeXML(f.get(o)+""));
				}
					//NB: if neither String nor primitive, it just falls right through
			}
		} catch (Exception e) {
			System.out.println("Field: "+e);
		}
	}

	/** This should likewise be in ParserXML.
		*/
	private static void printAttr(String attrName, String value) {
		//out.print(SPACER +attrName+ " = \"" +value+ "\"");
		out.print(" "+ attrName+ "=\"" +value+ "\"");
	}

	/** And this too (should be in ParserXML).
		*/
	private static String encodeXML(String in) {
		String a = "";	//output
		for (int i=0; i<in.length(); i++)
			switch (in.charAt(i)) {
				case '&': a+= "&amp;"; break;
				case '<': a+= "&lt;"; break;
				case '>': a+= "&gt;"; break;
				case '\'': a+= "&apos;"; break;
				case '"': a+= "&quot;"; break;
				default: a+= in.charAt(i);
			}
		return a;
	}


	/**--------------------------------------------------------------------------
		* References (map to Elements)
		* ----------
		**/
	//should factor out the two iterations over fields (attr and ref)
	private static void serializeReferences(Object o, String indent) {
		if (o==null) return;
		Field[] f = o.getClass().getFields();
			//Field[] f = o.getClass().getDeclaredFields(); //skips inherited fields
		for (int i=0; i<f.length; i++) {
			Class type = f[i].getType();
			try {
				if (!(type.isPrimitive() || type==String.class))		//.getName().equals("java.lang.String")))
					serializeObject(f[i].get(o), f[i].getName(), indent+SPACER);
			} catch (Exception e) {System.err.println("References: " +e);}
		}
	}



	/**--------------------------------------------------------------------------
		* Vector (container type)
		* ------
		* Contains objects only (primitives written as Wrappers inc String)
		* spc used instead of indent - is there any reason for this?
		**/
	private static void serializeVectorReferences (Vector o, String indent) {
		final String spc = indent + SPACER;
		if (o==null) throw new RuntimeException("impossible; redundant safety");
		for (Enumeration keys=o.elements(); keys.hasMoreElements(); ) {
			Object obj = keys.nextElement();
			serializeWrapperOrObject(obj, spc);
			//if (keys.hasMoreElements()) out.println();		//separating blank line
		}
	};		//what's the idea with the trailing ";"?



	/**--------------------------------------------------------------------------
		* Wrapper OR ObjectVector
		* -----------------------
		* A helper method (used for Vector and Array only so far), where the
		* object is assumed to have no "field name".
		**/
	private static void serializeWrapperOrObject (Object obj, String spc) {
		if (obj==null) {
			printOpenTag(NULL_TOKEN, spc);
			out.println("/>");		//good to have control of the newline here
			return;
		}
		Class type = obj.getClass();
		if (type==Boolean.class ||		//is there a cleverer way to do this?
				type==Byte.class ||				//we could loop, yes, but...
				type==Character.class ||  //wrappers !isPrimitive()
				type==Short.class ||
				type==Integer.class ||
				type==Long.class ||
				type==Float.class ||
				type==Double.class)
			serializeWrapper(obj, spc);
		else if (type==String.class)
			serializeWrapper(encodeXML((String)obj), spc);
		else
			serializeObject(obj, null, spc);	//null== no name
	}

	/**--------------------------------------------------------------------------
		* Hashtable (container type)
		* ---------
		* Each key and value is an object.
		* distinguished by position alone.
		* primitives type written as special "wrapper" objects (String, too).
		**/
	private static void serializeHashtableReferences (
			Hashtable o, String indent) {
		if (o==null) throw new RuntimeException("impossible; redundant safety");
		for (Enumeration keys=o.keys(); keys.hasMoreElements(); ) {
			Object key = keys.nextElement();
			Object value = o.get(key);
			Object[] objs = {key,        value};
			String[] spc =  {indent+SPACER, indent+SPACER};	//so can stagger, if we want
			for (int i=0; i<objs.length; i++) {
				Class type = objs[i].getClass();
				if (type==Boolean.class ||		//is there a cleverer way to do this?
						type==Byte.class ||				//we could loop, yes, but...
						type==Character.class ||  //wrappers !isPrimitive()
						type==Short.class ||
						type==Integer.class ||
						type==Long.class ||
						type==Float.class ||
						type==Double.class)
					serializeWrapper(objs[i], spc[i]);
				else if (type==String.class)
					serializeWrapper(encodeXML((String)objs[i]), spc[i]);
				else
					serializeObject(objs[i], null, spc[i]);	//null== no name
			}
			if (keys.hasMoreElements()) out.println();		//separating blank line
		}
	}

	/** We know it must be a String in this case; but how about a more general
		*	wrapper? 
		*
		**/
	private static void serializeWrapper(Object value, String indent) {
		String tag = value.getClass().getName();		//"java.lang.String";
		printOpenTag(tag, indent);
		printAttr("valueOf", value+"");
		out.print("/>");		//good to have control of the newline here
			//move into ParserXML?
		//printCloseTag(tag, "");
		//NOTE: a wrapper *always* is an empty tag, because it can't have children
		out.println();	//instead of printCloseTag()
	}


	/**--------------------------------------------------------------------------
		* Test class
		* ----------
		**/
	public static class Test {
		public static void main(String[] args) throws Exception {

			ChatMessage m;
			m = new Sound("me", "you", 123456789);
			//XMLSerialize.serialize(m);
			m = new Text("you", "me", "hello, anyone there?");
			//XMLSerialize.serialize(m);

			//XMLSerialize.serialize(new PrimitivesTest());
				//Q: how do you jump out of the namespace; ie have another "Test"?
	
			MyTyreClass eg;
			eg = new MyTyreClass(
					"<My car's \"Wheels\"> & info",
				new Wheel("Michelin",  new MyDate(31,10,1997)),
				new Wheel("Michelin",  new MyDate(31,10,1997)),
				new Wheel("Michelin",  new MyDate(2,3,1999)),
				new Wheel("Michelin",  new MyDate(14,9,2000))
			);
//			//XMLSerialize.serialize(eg);		//so simple to use!

		//HASHTABLE DEMO
			Hashtable h = new Hashtable();
			h.put("boolean", new Boolean(true));
			h.put("byte", new Byte((byte)8));
/*
			h.put("char", new Character((char)16));
			h.put("short", new Short((short)16));
			h.put("int", new Integer((int)32));	//redundant
			h.put("long", new Long((long)64));
			h.put("float", new Float((float)32));
			h.put("double", new Double((double)64));	//redundant
			h.put("String", "hey, wait, wait, String isn't a primitive! Or is it?");
*/
			//h.put("NULL Test 1", null);		//causes error
			//h.put(null, "NULL Test 2");		//causes error


	//ARRAY DEMO - I have commented out several, to make for a simple first demo
		//please try the others if you are curious
		//PRIMITIVE TYPES:
			int[][] ii = new int[2][2];
			ii[0][0] = 3;
			ii[0][1] = 9;
			ii[1][0] = 27;
			ii[1][1] = 81;
			XMLSerialize.serialize(ii);

			//XMLSerialize.serialize(new int[] {3,9,27,81});
			//boolean[][][] bb = new boolean[2][2][2];
			//bb[1][1][1] = true;
			//XMLSerialize.serialize(bb);
	
		//REFERENCE TYPES:
			//XMLSerialize.serialize(Array.newInstance(int.class, new int[] {1}));
//			XMLSerialize.serialize(new String[] {"red", "green", "blue"});
//			XMLSerialize.serialize(new MyDate[2] );	//empty array
/*
			MyDate d = new MyDate();	
			XMLSerialize.serialize(new MyDate[] {d,d} );

			ChatMessage[][][][] c = new ChatMessage[1][2][3][4];
			c[0][0][0][0] = new ChatMessage();
			XMLSerialize.serialize(c);
*/


		//VECTOR DEMO (uses above Hashtable)
			Vector v = new Vector();
			for (Enumeration keys=h.keys(); keys.hasMoreElements();) //above hash
				v.addElement(keys.nextElement());	//adds just the keys
			for (Enumeration keys=h.elements(); keys.hasMoreElements();) //above hash
			{
				v.addElement(keys.nextElement());	//adds just the keys
			//	System.err.println("\"" +keys.nextElement() + "\"");
			}
			//v.add(eg);
			//v.add(m);
			v.addElement(null);		//NOTE: this is a valid Vector element
			v.addElement("NOTE: null is a valid Vector element");
			XMLSerialize.serialize(v);		//so simple to use!
			
			//new Null();	  //this is in fact legal...

			try {
				PrintWriter tt =
					new PrintWriter(new BufferedWriter(new FileWriter("tt.xml")), true);
				XMLSerialize.serialize(h, tt);		//so simple to use!
				tt.close();
			} catch (Exception e) {System.err.println("IO: " +e);}
		}
	}
}
