/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the CCL-ZIP application.
 * (http://www.kclee.com/clemens/java/ccl/).
 *
 * The Initial Developer of the Original Code is Chr. Clemens Lee.
 * Portions created by Chr. Clemens Lee are Copyright (C) 2000
 * Chr. Clemens Lee. All Rights Reserved.
 *
 * Contributor(s): Chr. Clemens Lee <clemens@kclee.com>
 */

package test;

import ccl.util.FileUtil;
import ccl.util.Test;
import ccl.util.Util;
import zip;

/**
 * Self test of zip application.
 *
 * @version $Id: zipTest.java,v 1.3 2000/05/14 16:07:19 clemens Exp clemens $
 * @author  Chr. Clemens Lee clemens@kclee.com
 */
public class zipTest extends Test {
    /**
     * Get temporary test zip file name. Deletes also a file
     * in case it exists already.
     */
    private String _deleteAndGetTestZipFileName() {
        Util.debug( this, "_testDelete().tempDir: " + FileUtil.getTempDir() );
        String sTestZipFile = FileUtil.getTempDir();
        sTestZipFile = FileUtil.concatPath( sTestZipFile, 
                                            "zipTest.zip" );
        bugIf( FileUtil.delete( sTestZipFile ),
               "Could not delete '" + sTestZipFile + "'!" );

        return sTestZipFile;
    }

    /**
     * Get test file name number n.
     */
    private String _getFileName( String sDir_, int number_ ) {
        return FileUtil.concatPath
               ( sDir_, "zipentry" + number_ + ".txt" );
    }

    /**
     * Create two files to use for zipping.
     *
     * @exception   Exception   gets forwarded to ccl.util.Test
     *                          parent.
     */
    private void _createTwoFiles( String sDir_ ) 
        throws Exception
    {
        String sFirstFile  = _getFileName( sDir_, 1 );
        String sSecondFile = _getFileName( sDir_, 2 );
        FileUtil.writeFile( sFirstFile , "bla\n" );
        FileUtil.writeFile( sSecondFile, "blu\n" );
        Util.debug( this, "_testDelete().sSecondFile: " + sSecondFile );
    }

    /**
     * Test -d option.
     *
     * @exception   Exception   gets forwarded to ccl.util.Test
     *                          parent to count this test as
     *                          failed.
     */
    private void _testDelete() 
        throws Exception
    {
        String sTestZipFile = _deleteAndGetTestZipFileName();

        // create two files to use for zipping
        _createTwoFiles( FileUtil.getTempDir() );

        // zip these two files
        //    redirect stdout
        //    use std output for test results 
        Test.redirectStandardStreams( true );
        String[] asArg = { sTestZipFile
                           , _getFileName( FileUtil.getTempDir(), 1 )
                           , _getFileName( FileUtil.getTempDir(), 2 ) };
        new zip( asArg );
        bugIf( !FileUtil.existsFile( sTestZipFile ),
               "Test zip file '"
               + sTestZipFile
               + "' was not created!" );

        // delete one of these two files in zip file
        asArg = new String[] { "-d"
                               , sTestZipFile
                               , _getFileName( FileUtil.getTempDir(), 1 ) };
        new zip( asArg );
        Test.redirectStandardStreams( false );

        // check number of files in zip file
        // well, if there were a problem, an exception would
        // have been thrown
        bugIf( false );

        // delete all test files
        FileUtil.delete( sTestZipFile );
        FileUtil.delete( _getFileName( FileUtil.getTempDir(), 1 ) );
        FileUtil.delete( _getFileName( FileUtil.getTempDir(), 2 ) );
    }

    /**
     * Test -r option.
     *
     * @exception   Exception   gets forwarded to ccl.util.Test
     *                          parent to count this test as
     *                          failed.
     */
    private void _testRecurse() 
        throws Exception
    {
        String sTestZipFile = _deleteAndGetTestZipFileName();

        // create tmp directory
        String sTempZipDir = FileUtil.getTempDir();
        sTempZipDir = FileUtil.concatPath( sTempZipDir, "tmpziptest" );
        bugIf( FileUtil.md( sTempZipDir ), "Mkdir '" + sTempZipDir + "' failed!" );

        // create two files to use for zipping
        _createTwoFiles( sTempZipDir );

        // zip this directory
        //    redirect stdout
        //    use std output for test results 
        Test.redirectStandardStreams( true );
        String[] asArg = { "-r"
                           , sTestZipFile
                           , sTempZipDir };
        new zip( asArg );
        bugIf( !FileUtil.existsFile( sTestZipFile ),
               "Test zip file '"
               + sTestZipFile
               + "' was not created!" );
        Test.redirectStandardStreams( false );

        // well, if there were a problem, an exception would
        // have been thrown
        bugIf( false );

        // delete all test files
        FileUtil.delete( sTestZipFile );
        FileUtil.delete( _getFileName( sTempZipDir, 1 ) );
        FileUtil.delete( _getFileName( sTempZipDir, 2 ) );
        FileUtil.delete( sTempZipDir );
    }

    /**
     * Default Test constructor.
     */
    public zipTest() {
        super();
    }

    /**
     * Default Test constructor.
     */
    public zipTest( Test pTest_ ) {
        super( pTest_ );
    }

    /**
     * All the tests happen here. This gets invoked by the
     * test harness.
     *
     * @exception   Exception   gets forwarded to ccl.util.Test
     *                          parent to count this test as
     *                          failed.
     */
    protected void _doIt() 
        throws Exception 
    {
        _testDelete ();
        _testRecurse();
    }

    /**
     * Invoke zip selftest and exit.
     */
    public static void main( String[] asArg_ ) {
        Test pTest = (Test)(new zipTest());
        pTest.setVerbose( true );
        pTest.run();
        pTest.printResult( pTest );

        System.exit( 0 );
    }
}
