/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the CCL-ZIP application.
 * (http://www.kclee.com/clemens/java/ccl/).
 *
 * The Initial Developer of the Original Code is Chr. Clemens Lee.
 * Portions created by Chr. Clemens Lee are Copyright (C) 2000
 * Chr. Clemens Lee. All Rights Reserved.
 *
 * Contributor(s): Chr. Clemens Lee <clemens@kclee.com>
 */

import java.io.*;
import java.util.*;
import java.util.zip.*;

import ccl.util.*;
import test.zipTest;

/**
 * CCL-ZIP is a small Info-ZIP clone written in Java.
 *
 * @author  <a href="http://www.kclee.com/clemens/">Chr. Clemens Lahme</a> (<a href="mailto:clemens@kclee.com"><i>clemens@kclee.com</i></a>)
 * @version "$Header: /home/clemens/src/java/jacob/src/RCS/zip.java,v 2.9 2000/07/30 15:30:10 clemens Exp clemens $";
 */
public class zip implements Exitable {
    public static final String S_RCS_HEADER = "$Header: /home/clemens/src/java/jacob/src/RCS/zip.java,v 2.9 2000/07/30 15:30:10 clemens Exp clemens $";

    private static final String S_INI_FILE_CONTENT = 
          "[Init]\n"
        + "Author=Chr. Clemens Lee\n"
        + "\n"
        + "[Help]\n"
        + "; Please do not edit the Help section\n"
        + "HelpUsage=zipfile files\n"
        + "Options=j,d,e,r,check\n"
        + "j=b,o,junk (don't record) directory names\n"
        + "d=b,o,delete entries in zipfile\n"
        + "e=s,o,insErt entries with directory 'string' (store files with specified path)\n"
        + "r=b,o,recurse into directories\n"
        + "check=b,o,trigger zip selftests\n";

    private boolean _bExit = false;

    private byte[] _readZipEntry(ZipFile zipOpen_, 
                                 String sZipEntry_)
        throws IOException, ZipException
    {
        ZipEntry pZipEntry = zipOpen_.getEntry(sZipEntry_);
        int compressedSize = (int)pZipEntry.getCompressedSize();
        int uncompressedSize = (int)pZipEntry.getSize();
        if (uncompressedSize == -1) {
            throw new ZipException();
        }
        byte abFile[] = new byte[uncompressedSize];
        InputStream pInputStream = zipOpen_.getInputStream(pZipEntry);
        
        int bytesRead = 0;
        while(bytesRead < uncompressedSize) {
            bytesRead += pInputStream.read(abFile, bytesRead, uncompressedSize - bytesRead);
        }
        
        return abFile;
    }

    private void _insertFile(ZipOutputStream pZipOutputStream_,
                             Vector vFiles_, Vector vRealFiles_,
                             String sMethod_)
        throws FileNotFoundException, IOException
    {
        Enumeration eRealFiles = vRealFiles_.elements();
        for(Enumeration eFiles = vFiles_.elements();
            eFiles.hasMoreElements(); ) 
        {
            String sNextFileName = (String)eFiles.nextElement();
            String sNextRealFileName = (String)eRealFiles.nextElement();
            
            Util.println(sMethod_ + ": " + sNextFileName);
            
            if (FileUtil.existsDir(sNextRealFileName)) {
                continue;
            }
            FileInputStream pFileInputStream = new FileInputStream(sNextRealFileName);
            byte[] abFile = new byte[pFileInputStream.available()];
            pFileInputStream.read(abFile);
            pFileInputStream.close();
            
            ZipEntry pZipEntry = new ZipEntry(sNextFileName);
            pZipOutputStream_.putNextEntry(pZipEntry);
            
            pZipOutputStream_.write(abFile, 0, abFile.length);
            abFile = null;
            pZipOutputStream_.closeEntry();
        }
    }

    /**
     * For interactive usage. If you want to use this class
     * out of another class, just create a new zip object
     * and the zip file will be immediately created.
     *
     * The main routine also has a System.exit( 0 );
     */
    public zip( String[] asArg_ ) {
        super();

        Init pInit = new Init(this, 
                              asArg_, 
                              S_RCS_HEADER,
                              S_INI_FILE_CONTENT);
        if (_bExit) {
            return;
        }

        Util.debug( this, "<init>(..).asArg_: " +
                    Util.objectsToVector( asArg_ ) );
        
        Hashtable htOptions = pInit.getOptions();

        if ( htOptions.get( "check" ) != null ) {
            Test test = new zipTest();
            test.setVerbose( true );
            test.run();
            test.printResult( test );
            
            return;
        }

        Vector vNewRealFiles = pInit.getArguments();
        if (vNewRealFiles.size() < 2) {
            pInit.printHelpMessage();
            
            return;
        }
        
        boolean bDelete = (htOptions.get("d") != null);
        Util.debug("zip.<init>(..).bDelete: " + bDelete);
        
        String sZipFileName = (String)vNewRealFiles.elementAt(0);
        Util.panicIf( Util.isEmpty( sZipFileName ) );
        vNewRealFiles.removeElementAt(0);
        
        if (!Util.endsWith(sZipFileName, ".zip") &&
            !Util.endsWith(sZipFileName, ".jar"))
        {
            sZipFileName += ".zip";
        }
        String sTempFileName = null;
        if (!FileUtil.exists(sZipFileName)) {
            sTempFileName = sZipFileName;
        } else {
            sTempFileName = FileUtil.getTempFileName();
            if (sTempFileName == null) {
                Util.println("zip error: unable to find a temporary file name!");
                
                return;
            }
        }
        
        //
        // change/junk file names
        //
        String sNewDirectory = (String)htOptions.get("e");
        if (sNewDirectory != null &&
            sNewDirectory.charAt(sNewDirectory.length() - 1) != '/')
        {
            sNewDirectory += "/";
        }
        if (htOptions.get("j") != null) {
            sNewDirectory = "";
        }
        if ( sNewDirectory == null ) {
            sNewDirectory = "";
        }
	sNewDirectory = sNewDirectory.replace( File.separatorChar, '/' );
        Util.debug( "<init>(..).sNewDirectory: --->" + sNewDirectory + "<---" );

        boolean bRecursive = (htOptions.get( "r" ) != null);

        // there are two vectors of files. The real files on
        // the file system that shall be zipped: 
        //  - vNewRealFiles
        //
        // Then there is a second vector with the file names
        // under which they shall be stored inside the zip
        // archive:
        //  - vNewFiles
        Vector vNewFiles = new Vector();
        for( int file = 0; file < vNewRealFiles.size();
             file++ )
        {
            String sNextFileName = (String)vNewRealFiles.elementAt( file );
            Vector vsDir = null;
            if ( bRecursive && FileUtil.existsDir( sNextFileName ) ) {
                vsDir = FileUtil.getRecursiveDir( sNextFileName );
            }
            sNextFileName = sNextFileName.replace(File.separatorChar, '/');
            int offset = 0;
            if ( !Util.isEmpty( sNewDirectory ) ) {
                offset = sNextFileName.lastIndexOf('/') + 1; 
                sNextFileName = sNextFileName.substring
                       ( offset );
                sNextFileName = sNewDirectory + sNextFileName;
            }
            // if new file name is something like './zip.class'
            // then remove the leading './'.
            if ( sNextFileName.startsWith( "./" ) ) {
                sNextFileName = sNextFileName.substring( 2 );
            }
            vNewFiles.addElement(sNextFileName);
            if ( vsDir != null ) {
                for( int subfiles = 1; subfiles < vsDir.size(); subfiles++ ) {
                    String sNextSubFile = (String)vsDir.elementAt( subfiles );
                    vNewRealFiles.insertElementAt( sNextSubFile, file + subfiles );
                    sNextSubFile = sNextSubFile.replace( File.separatorChar, '/' );
                    if ( sNewDirectory != null ) {
                        sNextSubFile = sNextSubFile.substring( offset );
                        sNextSubFile = sNewDirectory + sNextSubFile;
                    }
                    vNewFiles.addElement( sNextSubFile );
                }
                file += vsDir.size() - 1;
            }
        }
        
        try {
            Vector vOriginalFiles = new Vector();
            Vector vUpdateFiles = new Vector();
            Vector vUpdateRealFiles = new Vector();
            ZipFile pZipFile = null;
            if (!sZipFileName.equals(sTempFileName)) {
                pZipFile = new ZipFile(sZipFileName);
                for(Enumeration eEntries = pZipFile.entries();
                    eEntries.hasMoreElements(); )
                {
                    String sEntry = ((ZipEntry)(eEntries.nextElement())).getName();
                    if (vNewFiles.contains(sEntry)) {
                        if (bDelete) {
                            Util.println("deleting: " + sEntry);
                        }
                        int index = vNewFiles.indexOf(sEntry);
                        String sRealFile = (String)vNewRealFiles.elementAt(index);
                        vUpdateFiles.addElement(sEntry);
                        vUpdateRealFiles.addElement(sRealFile);
                        vNewFiles.removeElementAt(index);
                        vNewRealFiles.removeElementAt(index);
                    } else {
                        vOriginalFiles.addElement(sEntry);
                    }
                }
            }

            FileOutputStream pFileOutputStream = new FileOutputStream(sTempFileName);
            BufferedOutputStream pBufferedOutputStream = new BufferedOutputStream
                   (pFileOutputStream);
            ZipOutputStream pZipOutputStream = new ZipOutputStream
                   (pBufferedOutputStream);
            
            // process original files
            for(Enumeration eOriginal = vOriginalFiles.elements();
                eOriginal.hasMoreElements(); )
            {
                String sNextFileName = (String)eOriginal.nextElement();
                Util.debug("zip.<ini>(..).sNextFileName: " + sNextFileName);
                byte[] abFile = _readZipEntry(pZipFile, sNextFileName);
                
                ZipEntry pZipEntry = new ZipEntry(sNextFileName);
                pZipOutputStream.putNextEntry(pZipEntry);
                
                pZipOutputStream.write(abFile, 0, abFile.length);
                abFile = null;
                pZipOutputStream.closeEntry();
            }
            
            if (!bDelete) {
		// updating
                _insertFile(pZipOutputStream, vUpdateFiles, vUpdateRealFiles, "updating");
                
		// adding
                _insertFile(pZipOutputStream, vNewFiles, vNewRealFiles, "  adding");
            }
            
            pZipOutputStream.close();
            if ( pZipFile != null ) {
                pZipFile.close();
            }

            if (!sTempFileName.equals(sZipFileName)) {
                FileUtil.move(sTempFileName, sZipFileName);
            }
        } catch(Exception pException) {
            Util.println("zip error: " + pException);
            try {
                if (!sTempFileName.equals(sZipFileName)) {
                    FileUtil.delete(sTempFileName);
                }
            } catch(Exception eInner) {
            }
        }
    }

    /**
     * Implements the ccl.util.Exitable interface.
     */
    public void setExit() {
        _bExit = true;
    }

    /**
     * For interactive usage. If you want to use this class
     * out of another class, just create a new zip object
     * and the zip file will be immediately created.
     *
     * The main routine also has a System.exit( 0 );
     */
    public static void main(String[] asArg_) {
        zip pzip = new zip
               ( asArg_ );
        
        System.exit(0);
    }
}
